@@ -1,10 +1,12 @@
+import re
+
 from cStringIO import StringIO
 
 from jinja2.visitor import NodeVisitor
 import jinja2.nodes
 import jinja2.compiler
 import jinja2.ext
-from jinja2.utils import Markup, concat, escape, is_python_keyword, next
+from jinja2.utils import escape, next
 
 import nodes
 
@@ -49,15 +51,6 @@ OPERATORS = {
     }
 
 
-def generate(node, environment, name, filename):
-    """Generate the python source for a node tree."""
-    if not isinstance(node, jinja2.nodes.Template):
-        raise TypeError("Can't compile non template nodes")
-    generator = CodeGenerator(environment, name, filename)
-    generator.visit(node)
-    return generator.writer.stream.getvalue()
-
-
 class JSFrameIdentifierVisitor(jinja2.compiler.FrameIdentifierVisitor):
 
     def __init__(self, identifiers, environment, ctx):
@@ -175,7 +168,7 @@ class JSFrame(jinja2.compiler.Frame):
 
 class StringBuilder(object):
 
-    def __init__(self):
+    def __init__(self, environment=None):
         self.stream = StringIO()
 
         # the current line number
@@ -194,9 +187,12 @@ class StringBuilder(object):
         # true if nothing was written so far.
         self._first_write = True
 
-        # the current indentation
+        # the current indentation multiplier
         self._indentation = 0
 
+        # the character(s) to display as a single indent
+        self._indentation_text = getattr(environment, 'js_indentation', '    ')
+
     # Copied
     def indent(self):
         """Indent by one."""
@@ -220,7 +216,7 @@ class StringBuilder(object):
                                             self.code_lineno))
                     self._write_debug_info = None
             self._first_write = False
-            self.stream.write('    ' * self._indentation)
+            self.stream.write(self._indentation_text * self._indentation)
             self._new_lines = 0
         self.stream.write(x)
 
@@ -257,8 +253,9 @@ class StringBuilder(object):
     # output formating
 
     def writeline_startoutput(self, node, frame):
-        self.writeline("var output = %s_sb || new goog.string.StringBuffer();" %(
-            frame.parameter_prefix))
+        self.writeline(
+            "var output = %s_sb || new goog.string.StringBuffer();" %(
+                frame.parameter_prefix))
 
     def writeline_endoutput(self, node, frame):
         self.writeline(
@@ -282,15 +279,13 @@ class StringBuilder(object):
 
 class Concat(StringBuilder):
 
-    def __init__(self):
-        super(Concat, self).__init__()
-
     def writeline_provides(self, node, frame, namespace):
         parts = namespace.split(".")
         for idx, part in enumerate(parts):
             ns = ".".join(parts[:idx + 1])
-            self.writeline("if (typeof %s == 'undefined') { %s%s = {}; }" %(
-                ns, idx == 0 and "var " or "", ns), node)
+            self.writeline(
+                "if (typeof %s == 'undefined') { %s%s = {}; }" %(
+                    ns, idx == 0 and "var " or "", ns), node)
 
     def writeline_require(self, node, frame, namespace):
         self.newline(node)
@@ -333,14 +328,6 @@ class BaseCodeGenerator(NodeVisitor):
         self.name = name
         self.filename = filename
 
-        self.writer = environment.writer()
-
-    # Copied
-    def fail(self, msg, lineno):
-        """Fail with a :exc:`TemplateAssertionError`."""
-        raise jinja2.compiler.TemplateAssertionError(
-            msg, lineno, self.name, self.filename)
-
     def blockvisit(self, nodes, frame):
         """
         Visit a list of noes ad block in a frame. Some times we want to
@@ -361,7 +348,9 @@ class CodeGenerator(BaseCodeGenerator):
         """
         namespace = list(node.find_all(nodes.NamespaceNode))
         if len(namespace) > 1:
-            self.fail("You can only supply one namespace per template", 0)
+            raise jinja2.compiler.TemplateAssertionError(
+                "You can only supply one namespace per template",
+                0, self.name, self.filename)
         if namespace:
             namespace = namespace[0].namespace
         else:
@@ -400,23 +389,32 @@ class CodeGenerator(BaseCodeGenerator):
     def visit_Macro(self, node, frame):
         generator = MacroCodeGenerator(
             self.environment,
+            self.writer.__class__(self.environment),
             self.name,
             self.filename)
         generator.visit(node, frame)
 
-        for requirement in generator.requirements:
-            self.writer.writeline_require(
-                node, frame, requirement.encode(frame.eval_ctx.encoding))
-        if generator.requirements:
-            self.writer.newline(node) # keep whitespace ok
-
         self.writer.write(generator.writer.stream.getvalue())
 
     def visit_TemplateData(self, node, frame):
         self.writer.mark(node)
         self.writer.write(node.data)
-        if node.data[-1] != "\n":
-            self.writer.newline()
+
+
+class ClosureCodeGenerator(CodeGenerator):
+
+    def __init__(self, environment, name, filename):
+        super(ClosureCodeGenerator, self).__init__(environment, name, filename)
+
+        self.writer = StringBuilder()
+
+
+class ConcatCodeGenerator(CodeGenerator):
+
+    def __init__(self, environment, name, filename):
+        super(ConcatCodeGenerator, self).__init__(environment, name, filename)
+
+        self.writer = Concat()
 
 
 STRINGBUILDER = "StringBuilder"
@@ -432,17 +430,10 @@ class MacroCodeGenerator(BaseCodeGenerator):
     # comments should be displayed in the JS file. We need them for any closure
     # compiler hints we may want to put in.
 
-    def __init__(self, environment, name, filename):
+    def __init__(self, environment, writer, name, filename):
         super(MacroCodeGenerator, self).__init__(environment, name, filename)
 
-        # collect all the namespaced requirements
-        self.requirements = set([])
-
-    def addRequirement(self, requirement, frame):
-        if requirement == frame.eval_ctx.namespace:
-            return
-
-        self.requirements.add(requirement)
+        self.writer = writer
 
     def visit_Output(self, node, frame):
         # JS is only interested in macros etc, as all of JavaScript
@@ -492,19 +483,26 @@ class MacroCodeGenerator(BaseCodeGenerator):
                     start = False
                 else:
                     self.writer.write_outputappend_add(node, frame)
+                if getattr(self.environment, "strip_html_whitespace", False):
+                    item = [strip_html_whitespace(itemhtml)
+                            for itemhtml in item]
                 self.writer.write(repr("".join(item)))
             else:
+                # This is a non-data node.
                 # If we are using the string builder then we generate slightly
                 # different code then concat.
                 if self.writer.__class__.__name__ == STRINGBUILDER:
                     if isinstance(item, jinja2.nodes.Call):
-                        if not start:
-                            self.writer.write_outputappend_end(item, frame)
-                            start = True
-                        self.writer.newline(item)
-                        self.visit(item, frame)
-                        self.writer.write(";")
-                        continue
+                        # XXX - If we are a macro then we must end the
+                        # appending of the output.
+                        if not item.args:
+                            if not start:
+                                self.writer.write_outputappend_end(item, frame)
+                                start = True
+                            self.writer.newline(item)
+                            self.visit(item, frame)
+                            self.writer.write(";")
+                            continue
 
                 if start:
                     self.writer.writeline_outputappend(item, frame)
@@ -512,9 +510,9 @@ class MacroCodeGenerator(BaseCodeGenerator):
                 else:
                     self.writer.write_outputappend_add(item, frame)
 
-                # autoescape, safe
+                # autoescape, safe, and escape
                 if isinstance(item, jinja2.nodes.Filter):
-                    if frame.eval_ctx.autoescape and item.name == "safe":
+                    if item.name == "safe":
                         self.visit(item.node, frame)
                         continue
 
@@ -553,7 +551,9 @@ class MacroCodeGenerator(BaseCodeGenerator):
 
             FILTERS[node.name](self, node, frame, *node.args, **kwargs)
         else:
-            self.fail("Filter does not exist: '%s'" % node.name, node.lineno)
+            raise jinja2.compiler.TemplateAssertionError(
+                "Filter does not exist: '%s'" % node.name,
+                node.lineno, self.name, self.filename)
 
     def visit_Const(self, node, frame, dotted_name = None):
         # XXX - need to know the JavaScript ins and out here.
@@ -590,7 +590,7 @@ class MacroCodeGenerator(BaseCodeGenerator):
 
             # item.key should be a constant string. Otherwise how do you
             # get to output a dictionary with a variable key string?
-            # So it is either a `Const` or `Name` node. 
+            # So it is either a `Const` or `Name` node.
             if isinstance(item.key, jinja2.nodes.Const):
                 if "." in item.key.value:
                     raise jinja2.compiler.TemplateAssertionError(
@@ -650,8 +650,24 @@ class MacroCodeGenerator(BaseCodeGenerator):
             frame.assigned_names.add(frame.identifiers.imports[name])
         else:
             if dotted_name is None:
-                self.fail("Variable '%s' not defined" % name, node.lineno)
-            output = node.name
+                raise jinja2.compiler.TemplateAssertionError(
+                    "Variable '%s' not defined" % name,
+                    node.lineno, self.name, self.filename)
+
+            if not dotted_name and frame.eval_ctx.namespace:
+                # For compatibility with regular jinja2 env, when one macro
+                # calls itself or another macro within the same template, it
+                # will not specify the namespace in the macro name. For the
+                # compiled JS we must inject the namespace.
+                top_frame = frame
+                while top_frame.parent:
+                    top_frame = top_frame.parent
+
+                namespaced_name = frame.eval_ctx.namespace + '.' + node.name
+                if namespaced_name in top_frame.identifiers.declared_locally:
+                    name = namespaced_name
+
+            output = name
 
         if dotted_name is None:
             self.writer.write(output)
@@ -660,19 +676,33 @@ class MacroCodeGenerator(BaseCodeGenerator):
 
         return isparam
 
+    def visit_Getitem(self, node, frame):
+        # Careful, there is something in Jinja2 about node.arg extending
+        # jinja2.nodes.Slice
+        self.visit(node.node, frame)
+        self.writer.write("[")
+        self.visit(node.arg, frame)
+        self.writer.write("]")
+
     def visit_Getattr(self, node, frame, dotted_name = None):
-        if frame.forloop_buffer and node.node.name == "loop":
+        # We only need to check when `loop` is the first name-space in the
+        # Getattr node. Sometimes we have more then one level name-spaces and
+        # we are inside a for loop, likely when we are calling other macros.
+        if frame.forloop_buffer and getattr(node.node, "name", None) == "loop":
             if node.attr == "index0":
                 self.writer.write("%sIndex" % frame.forloop_buffer)
             elif node.attr == "index":
-                self.writer.write("%sIndex + 1" % frame.forloop_buffer)
+                self.writer.write("(%sIndex + 1)" % frame.forloop_buffer)
             elif node.attr == "revindex0":
-                self.writer.write("%sListLen - %sIndex" %(frame.forloop_buffer,
-                                                          frame.forloop_buffer))
+                self.writer.write("(%sListLen - %sIndex)" %(
+                    frame.forloop_buffer,
+                    frame.forloop_buffer)
+                    )
             elif node.attr == "revindex":
                 self.writer.write(
-                    "%sListLen - %sIndex - 1" %(frame.forloop_buffer,
-                                                frame.forloop_buffer))
+                    "(%sListLen - %sIndex - 1)" %(frame.forloop_buffer,
+                                                  frame.forloop_buffer)
+                    )
             elif node.attr == "first":
                 self.writer.write("%sIndex == 0" % frame.forloop_buffer)
             elif node.attr == "last":
@@ -684,6 +714,9 @@ class MacroCodeGenerator(BaseCodeGenerator):
             else:
                 raise AttributeError("loop.%s not defined" % node.attr)
         else:
+            # write_variable is going to be true if dotted_name is None which
+            # implies that we are gathering the variable name together. So
+            # don't write it out yet.
             write_variable = False
             if dotted_name is None:
                 dotted_name = []
@@ -694,8 +727,6 @@ class MacroCodeGenerator(BaseCodeGenerator):
             dotted_name.append(node.attr)
 
             if write_variable:
-                if not param:
-                    self.addRequirement(".".join(dotted_name[:-1]), frame)
                 self.writer.write(".".join(dotted_name))
 
     def binop(operator):
@@ -755,9 +786,9 @@ class MacroCodeGenerator(BaseCodeGenerator):
 
     def visit_Operand(self, node, frame):
         if node.op not in OPERATORS:
-            self.fail(
+            raise jinja2.compiler.TemplateAssertionError(
                 "Comparison operator '%s' not supported in JavaScript",
-                node.lineno)
+                node.lineno, self.name, self.filename)
         self.writer.write(" %s " % OPERATORS[node.op])
         self.visit(node.expr, frame)
 
@@ -799,8 +830,9 @@ class MacroCodeGenerator(BaseCodeGenerator):
             loop_frame.forloop_buffer = node.target.name
         for name in node.find_all(jinja2.nodes.Name):
             if name.ctx == "store" and name.name == "loop":
-                self.fail("Can't assign to special loop variable "
-                          "in for-loop target", name.lineno)
+                raise jinja2.compiler.TemplateAssertionError(
+                    "Can't assign to special loop variable in for-loop target",
+                    name.lineno, self.name, self.filename)
 
         self.writer.writeline("var %sList = " % node.target.name)
         self.visit(node.iter, loop_frame)
@@ -855,9 +887,12 @@ class MacroCodeGenerator(BaseCodeGenerator):
              func_frame.identifiers.declared_parameter)
         )
         if overriden_closure_vars:
-            self.fail("It's not possible to set and access variables "
-                      "derived from an outer scope! (affects: %s)" %
-                      ", ".join(sorted(overriden_closure_vars)), node.lineno)
+            raise jinja2.compiler.TemplateAssertionError(
+                "It's not possible to set and access variables "
+                "derived from an outer scope! (affects: %s)" %(
+                    ", ".join(sorted(overriden_closure_vars)), node.lineno),
+                lineno, self.name, self.filename
+                )
 
         # remove variables from a closure from the frame's undeclared
         # identifiers.
@@ -879,11 +914,12 @@ class MacroCodeGenerator(BaseCodeGenerator):
             node, frame, children = children,
             parameter_prefix = parameter_prefix)
 
-        self.writer.writeline("%s = function(%s_data, %s_sb, %s_caller) {" %(
-            name,
-            frame.parameter_prefix,
-            frame.parameter_prefix,
-            frame.parameter_prefix))
+        self.writer.writeline(
+            "%s = function(%s_data, %s_sb, %s_caller) {" %(
+                name,
+                frame.parameter_prefix,
+                frame.parameter_prefix,
+                frame.parameter_prefix))
         self.writer.indent()
         if node.defaults:
             # We have defaults to work with. Loop over all the default keys
@@ -904,23 +940,37 @@ class MacroCodeGenerator(BaseCodeGenerator):
                 start = False
             self.writer.write("};")
             self.writer.writeline("for (var key in defaults) {")
+            self.writer.indent()
             self.writer.writeline(
-                "    if (!(key in %s_data)) {" % frame.parameter_prefix)
+                "if (!(key in %s_data)) {" % frame.parameter_prefix)
+            self.writer.indent()
             self.writer.writeline(
-                "        %s_data[key] = defaults[key];" % frame.parameter_prefix)
-            self.writer.writeline("    }")
+                "%s_data[key] = defaults[key];" % (
+                    frame.parameter_prefix))
+            self.writer.outdent()
+            self.writer.writeline("}")
+            self.writer.outdent()
             self.writer.writeline("}")
         self.writer.writeline_startoutput(node, frame)
         self.blockvisit(node.body, frame)
         self.writer.writeline_endoutput(node, frame)
         self.writer.outdent()
-        self.writer.writeline("}")
+        self.writer.writeline("};")
 
     def visit_Macro(self, node, frame):
         name = node.name
         if frame.eval_ctx.namespace:
             name = frame.eval_ctx.namespace + "." + name
 
+        if getattr(self.environment, "add_compiler_annotations", False):
+            self.writer.writeline('/**')
+            self.writer.writeline(' * @param {Object.<string, *>=} opt_data')
+            self.writer.writeline(' * @param {goog.string.StringBuffer=} opt_sb')
+            self.writer.writeline(' * @param {Function=} opt_caller')
+            self.writer.writeline(' * @return {string|undefined}')
+            self.writer.writeline(' * @notypecheck')
+            self.writer.writeline(' */')
+
         body = self.macro_body(name, node, frame)
         frame.assigned_names.add("%s.%s" %(frame.eval_ctx.namespace, node.name))
 
@@ -941,41 +991,51 @@ class MacroCodeGenerator(BaseCodeGenerator):
             self.writer.writeline_outputappend(node, frame)
         else:
             # XXX - we shouldn't get here
-            self.fail("Unknown writer class", node.lineno)
+            raise jinja2.compiler.TemplateAssertionError(
+                "Unknown writer class", node.lineno, self.name, self.filename)
 
         self.visit(node.call, frame, forward_caller = "func_caller")
         self.writer.write(";")
 
-    def signature(self, node, frame):
-        if node.args:
-            self.fail(
-                "Function call with positional arguments not allowed with JS",
-                node.lineno)
+    def signature(self, node, frame, forward_caller):
+        if node.args and node.kwargs:
+            raise jinja2.compiler.TemplateAssertionError(
+                "Function call with positional and keyword arguments "
+                "are not allowed", node.lineno, self.name, self.filename)
 
         if node.dyn_args or node.dyn_kwargs:
-            self.fail(
+            raise jinja2.compiler.TemplateAssertionError(
                 "JS Does not support positional or keyword arguments",
-                node.lineno)
+                node.lineno, self.name, self.filename)
 
-        start = True
+        if node.args:
+            # We have only positional arguments here. In this case we assume
+            # that we have an ordinary Java Script function we wish to call
+            start_arg = True
+            for arg in node.args:
+                if not start_arg:
+                    self.writer.write(", ")
+                self.visit(arg, frame)
+                start_arg = False
+
+            # Since this is a ordinary call bail out
+            return
+
+        # Now assume that we are calling an other macro
+        # XXX - we should be able test this by looking up the environment to
+        # see if that is the case.
+
+        start_kw = True
         self.writer.write("{")
         for kwarg in node.kwargs:
-            if not start:
+            if not start_kw:
                 self.writer.write(", ")
-                start = False
             self.writer.write(kwarg.key)
             self.writer.write(": ")
             self.visit(kwarg.value, frame)
+            start_kw = False
         self.writer.write("}")
 
-    def visit_Call(self, node, frame, forward_caller = None):
-        # function symbol to call
-        dotted_name = []
-        self.visit(node.node, frame, dotted_name = dotted_name)
-        # function signature
-        self.writer.write("%s(" % ".".join(dotted_name))
-        self.signature(node, frame)
-
         # If we are using the string builder then we generate slightly
         # different code.
         if self.writer.__class__.__name__ == STRINGBUILDER:
@@ -989,8 +1049,61 @@ class MacroCodeGenerator(BaseCodeGenerator):
 
             self.writer.write(", ")
             self.writer.write(forward_caller)
+
+    def visit_Call(self, node, frame, forward_caller = None):
+        # function symbol to call
+        dotted_name = []
+        self.visit(node.node, frame, dotted_name = dotted_name)
+        func_name = ".".join(dotted_name)
+
+        # Like signature(), this assumes that function calls with only
+        # positional arguments are calls to javascript functions. This
+        # allows you to call differently named functions in the client JS
+        # than when rendering a template on the server.
+        if node.args and not node.kwargs \
+               and func_name in getattr(self.environment, "js_func_aliases", []):
+            func_name = self.environment.js_func_aliases[func_name]
+
+        # function signature
+        self.writer.write("%s(" % func_name)
+        self.signature(node, frame, forward_caller)
         self.writer.write(")")
 
+    def visit_Assign(self, node, frame):
+        # XXX - test that we don't override any variable names doing this
+        self.writer.newline(node)
+        self.writer.write("var ")
+        self.visit(node.target, frame)
+        self.writer.write(" = ")
+        self.visit(node.node, frame)
+        self.writer.write(";")
+
+    def visit_CondExpr(self, node, frame):
+        if self.writer.__class__.__name__ == CONCAT:
+            self.writer.write('(')
+        self.visit(node.test, frame)
+        self.writer.write(' ? ')
+        self.visit(node.expr1, frame)
+        self.writer.write(' : ')
+        # XXX - should this allow undefined else clauses? it's tough to
+        #       really judge what the correct behaviour should be, but
+        #       I have chosen an empty string since it is falsy and won't
+        #       be printed if that's what the value is ultimately used for.
+        if node.expr2:
+            self.visit(node.expr2, frame)
+        else:
+            self.writer.write("''")
+        if self.writer.__class__.__name__ == CONCAT:
+            self.writer.write(')')
+
+_pre_tag_whitespace = re.compile(r'\s*<')
+_post_tag_whitespace = re.compile(r'>\s*')
+_excess_whitespace = re.compile(r'\s\s+')
+
+def strip_html_whitespace(value):
+    value = _pre_tag_whitespace.sub('<', value)
+    value = _post_tag_whitespace.sub('>', value)
+    return _excess_whitespace.sub(' ', value)
 
 FILTERS = {}
 
@@ -1075,3 +1188,28 @@ def filter_round(generator, node, frame, precision = jinja2.nodes.Const(0)):
     generator.writer.write(")")
     if precision > 1:
         generator.writer.write(" / %s" % precision)
+
+
+def _generate(node, generator):
+    if not isinstance(node, jinja2.nodes.Template):
+        raise TypeError("Can't compile non template nodes")
+
+    generator.visit(node)
+    return generator.writer.stream.getvalue()
+
+
+def generate(node, environment, name, filename):
+    """Generate the python source for a node tree."""
+    generator = CodeGenerator(environment, name, filename)
+    generator.writer = environment.writer(environment)
+    return _generate(node, generator)
+
+
+def generateClosure(node, environment, name, filename):
+    generator = ClosureCodeGenerator(environment, name, filename)
+    return _generate(node, generator)
+
+
+def generateConcat(node, environment, name, filename):
+    generator = ConcatCodeGenerator(environment, name, filename)
+    return _generate(node, generator)
